import torch, scipy, scanpy, numpy, random
from typing import Literal
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import kneighbors_graph

def set_random_seed(seed: int = 0):
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    random.seed(seed)

def getX(adata, modality: Literal["rna", "atac", "protein", "metabolite"]):
    adata.var_names_make_unique()
    if modality in ['rna', 'atac']:
        scanpy.pp.filter_genes(adata, min_cells=10)
        scanpy.pp.log1p(adata)

        if scipy.sparse.issparse(adata.X):
            return adata.X.toarray()
        else:
            return adata.X

    elif modality=='protein':
        adata.X = numpy.apply_along_axis(protein_norm, 1, (adata.X.toarray() if scipy.sparse.issparse(adata.X) else numpy.array(adata.X)))
        return adata.X     

    elif modality=='metabolite':
        scanpy.pp.log1p(adata)
        if scipy.sparse.issparse(adata.X):
            return adata.X.toarray()
        else:
            return adata.X

def protein_norm(x):
    s = numpy.sum(numpy.log1p(x[x > 0]))
    exp = numpy.exp(s / len(x))
    return numpy.log1p(x / exp)

def cal_dist(X1, sparse: bool = False, neighbors: int = 100):
    D = pairwise_distances(X1)
    sig = numpy.median(D[numpy.triu_indices_from(D, k=1)])
    if not sparse:
        a1 = numpy.exp(-1*(D**2)/(2*(sig**2)))
        return a1
    else:
        dist1 = kneighbors_graph(X1, n_neighbors = neighbors, mode='distance')
        dist1.data = numpy.exp(-1*(dist1.data**2)/(2*(sig**2)))
        dist1.eliminate_zeros()
        return dist1



import torch, numpy, scanpy
from sklearn.preprocessing import StandardScaler
from .utils import cal_dist
from sklearn.decomposition import PCA
from scipy.sparse import coo_matrix
from itertools import combinations
from .nets import KAN
from typing import Union, List, Tuple, Literal

try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except NameError:
    from tqdm import tqdm

class mmspao(torch.nn.Module):
    def __init__(self, features, sel_views: Union[List[int], Literal["all"]]='all', combination: Union[List[Tuple[int, int]], Literal["all"]]='all', sparse: bool=False, neighbors: Union[int, None] = None, device: Literal["cpu", 'cuda']='cpu'):
        super(mmspao, self).__init__()
        self.device = device
        self.num_views = len(features) # numbers of modalities
        self.sparse = sparse # get from scipy.sparse.issparse(adata.X)
        if neighbors is None and self.sparse:
            neighbors = 100
        adj, pcs = [], []
        for i in tqdm(features, desc='perform PCA for each modality'):
            i = StandardScaler().fit_transform(i)
            adj.append(cal_dist(i, sparse = self.sparse, neighbors=neighbors))
            pcs.append(PCA(128).fit_transform(i) if i.shape[1] > 128 else i)
        self.pcs = [torch.Tensor(i).to(self.device) for i in pcs] 
        if not self.sparse:
            self.adj = [torch.Tensor(i).to(self.device) for i in adj]
        else:
            indices, values, shape = [], [], []
            for i in tqdm(adj, desc='sparse adj for each modality'):
                i = coo_matrix(i)
                indices.append(torch.LongTensor(numpy.vstack((i.row, i.col))))
                values.append(torch.FloatTensor(i.data))
                shape.append(torch.Size(i.shape))
            self.adj = [torch.sparse_coo_tensor(indices[i], values[i], shape[i]).to(self.device) for i in range(len(adj))]

        if sel_views == 'all':
            self.sel_views = list(range(len(self.pcs)))
        else:
            self.sel_views = sel_views   
        if combination == 'all':
            self.combinations = list(combinations(list(range(len(self.pcs))),2))
        else:
            self.combinations = combination        

    def train(self, epochs: int=1000):
        self.models = [KAN(pca_shape = self.pcs[i].shape[1], latent_reps = 64).to(self.device) for i in range(len(self.pcs))]
        def graph_emb(mod_adj, cur_emb):
            if not self.sparse:
                return (torch.triu(torch.cdist(cur_emb, cur_emb))*torch.triu(mod_adj)).mean()
            else:
                indices0 = mod_adj.coalesce().indices()[0]
                indices1 = mod_adj.coalesce().indices()[1]
                rows = cur_emb[indices0]
                cols = cur_emb[indices1]
                norm_dist = torch.norm(rows - cols, dim=1)
                return (norm_dist*mod_adj.coalesce().values()).mean()

        for i in range(self.num_views):
            self.models[i].train()
            optimizer = torch.optim.Adam(self.models[i].parameters(), lr=1e-3)
            pbar = tqdm(range(epochs), desc='Training KAN for modality ' + str(i+1))          
            for epoch in pbar:
                optimizer.zero_grad()
                gen = self.models[i](self.pcs[i])
                latent = self.models[i].get_latent(self.pcs[i])
                loss1 = torch.nn.MSELoss()(self.pcs[i], gen)
                loss2 = graph_emb(self.adj[i], latent)
                loss = loss1 + loss2
                pbar.set_postfix({'gen_loss' : '{:.3f}'.format(loss1),
                                  'sc_loss': '{:.3f}'.format(loss2),
                                  'loss': '{:.3f}'.format(loss)})
                loss.backward()
                optimizer.step()

        for i in range(self.num_views):
            self.models[i].eval()
        eval_embs = [self.models[i].get_latent(self.pcs[i]) for i in range(self.num_views)]
        if self.combinations is not None:
            exterior = [eval_embs[i][:, :, None]*eval_embs[j][:, None, :] for i, j in self.combinations]
            exterior = [i.reshape(i.shape[0], -1) for i in exterior]
            exterior = [torch.matmul(i, torch.pca_lowrank(i, q=64)[2]) for i in exterior]
        eval_embs = [eval_embs[i] for i in self.sel_views]
        eval_embs = [StandardScaler().fit_transform(i.cpu().detach().numpy()) for i in eval_embs]
        eval_embs = numpy.concatenate(eval_embs, 1)
        if self.combinations is not None:
            exterior = [StandardScaler().fit_transform(i.cpu().detach().numpy()) for i in exterior]
            exterior = numpy.concatenate(exterior, 1)
            emb = numpy.concatenate((eval_embs, exterior), 1)
        else:
            emb = eval_embs
        self.emb = emb
    
    def cluster(self, adata, n_domains: int, n_neighbors: int=50, end: float=2.5):
        assert end > 0.1
        adata.obsm['emb'] = self.emb
        scanpy.pp.neighbors(adata, n_neighbors=n_neighbors, use_rep='emb')
        pbar = tqdm(sorted(numpy.arange(0.1, end, 0.01), reverse=True), desc='Leiden clustering')
        for res in pbar:
            scanpy.tl.leiden(adata, random_state=0, resolution=res, flavor="igraph", n_iterations=2, directed=False)
            if len(adata.obs['leiden'].unique()) == n_domains:
                print(f"Found resolution: {res:.2f} for {n_domains} domains")
                return res
        return 1.0


from setuptools import setup
from mmspao import __version__

try:
    with open('README.md', 'r', encoding='utf-8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

setup(
      name='mmspao',  # pkg_name
      packages=['mmspao',],
      version=__version__,  # version number
      description="Multi-modal spatial omics integration with views and combinations.",
      author='林景',
      author_email='linjing010729@163.com',
      license='MIT',
      url='https://github.com/linjing-lab/mmspao',
      download_url='https://github.com/linjing-lab/mmspao/tags',
      long_description=_long_description,
      long_description_content_type='text/markdown',
      include_package_data=True,
      zip_safe=False,
      setup_requires=['setuptools>=18.0', 'wheel'],
      project_urls={
            'Source': 'https://github.com/linjing-lab/mmspao/tree/main/mmspao/',
            'Tracker': 'https://github.com/linjing-lab/mmspao/issues',
      },
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Healthcare Industry',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'License :: OSI Approved :: MIT License',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      install_requires=[
            'numpy>=1.26.4', # 'numpy==1.26.4'
            'scanpy>=1.10.4',
            'scikit-learn>=1.7.1',
            'scipy>=1.15.3',
            'tqdm>=4.67.1',
      ],
      # extras_require=[]
)